@@ -1,6 +1,6 @@
 /*****************************************************************************
  * Plus42 -- an enhanced HP-42S calculator simulator
- * Copyright (C) 2004-2022  Thomas Okken
+ * Copyright (C) 2004-2025  Thomas Okken
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License, version 2,
@@ -178,7 +178,7 @@ int docmd_comb(arg_struct *arg) {
     #ifdef BCD_MATH
         s = x == 0 ? 1 : pow(10, 1 + floor(log10(x)));
     #elif ANDROID
-        s = x == 0 ? 1 : pow(2, 1 + floor(log(x) / log(2)));
+        s = x == 0 ? 1 : pow(2, 1 + floor(log(x) / log(phloat(2))));
     #else
         s = x == 0 ? 1 : pow(2, 1 + floor(log2(x)));
     #endif
@@ -430,10 +430,6 @@ int docmd_input(arg_struct *arg) {
 
 int view_helper(arg_struct *arg, bool print) {
     int err;
-    char *buf = (char *) malloc(disp_c);
-    freer f(buf);
-    int bufptr = 0, part2;
-    vartype *v;
     if (arg->type == ARGTYPE_IND_NUM
             || arg->type == ARGTYPE_IND_STK
             || arg->type == ARGTYPE_IND_STR) {
@@ -441,41 +437,88 @@ int view_helper(arg_struct *arg, bool print) {
         if (err != ERR_NONE)
             return err;
     }
-    switch (arg->type) {
-        case ARGTYPE_NUM: {
-            int num = arg->val.num;
-            char2buf(buf, disp_c, &bufptr, 'R');
-            if (num < 10)
-                char2buf(buf, disp_c, &bufptr, '0');
-            bufptr += int2string(num, buf + bufptr, disp_c - bufptr);
-            break;
-        }
-        case ARGTYPE_STK:
-            string2buf(buf, disp_c, &bufptr, "ST ", 3);
-            char2buf(buf, disp_c, &bufptr, arg->val.stk);
-            break;
-        case ARGTYPE_STR:
-            string2buf(buf, disp_c, &bufptr, arg->val.text, arg->length);
-            break;
-    }
-    char2buf(buf, disp_c, &bufptr, '=');
-    part2 = bufptr;
+    vartype *v;
     err = generic_rcl(arg, &v);
     if (err != ERR_NONE)
         return err;
-    bufptr += vartype2string(v, buf + bufptr, disp_c - bufptr);
-    free_vartype(v);
-    draw_message(0, buf, bufptr);
 
+    char buf[64];
+    int bufptr = 0;
+    switch (arg->type) {
+        case ARGTYPE_NUM: {
+            int num = arg->val.num;
+            char2buf(buf, 64, &bufptr, 'R');
+            if (num < 10)
+                char2buf(buf, 64, &bufptr, '0');
+            bufptr += int2string(num, buf + bufptr, 64 - bufptr);
+            break;
+        }
+        case ARGTYPE_STK:
+            string2buf(buf, 64, &bufptr, "ST ", 3);
+            char2buf(buf, 64, &bufptr, arg->val.stk);
+            break;
+        case ARGTYPE_STR:
+            string2buf(buf, 64, &bufptr, arg->val.text, arg->length);
+            break;
+    }
+    char2buf(buf, 64, &bufptr, '=');
+    int part2 = bufptr;
+
+    int vlength;
+    if (v->type == TYPE_STRING) {
+        vartype_string *s = (vartype_string *) v;
+        vlength = s->length + 2;
+    } else if (v->type == TYPE_EQUATION) {
+        vartype_equation *eq = (vartype_equation *) v;
+        vlength = eq->data->length + 2;
+    } else {
+        vlength = 42; // max length of complex number
+    }
+    char *sbuf = NULL;
+    char *pbuf = buf;
+    int maxlen = disp_r * disp_c;
+    int slen = bufptr + vlength;
+    if (slen > maxlen)
+        slen = maxlen;
+    if (slen > 64) {
+        sbuf = (char *) malloc(slen);
+        if (sbuf == NULL) {
+            slen = 64;
+        } else {
+            memcpy(sbuf, buf, bufptr);
+            pbuf = sbuf;
+        }
+    }
+    bufptr += vartype2string(v, pbuf + bufptr, slen - bufptr);
+    draw_long_message(0, pbuf, bufptr);
+    free(sbuf);
+    bufptr = part2;
+
+    err = ERR_NONE;
     if (print && (flags.f.printer_enable || !program_running())) {
         if (flags.f.printer_exists) {
-            shell_annunciators(-1, -1, 1, -1, -1, -1);
-            print_wide(buf, part2, buf + part2, bufptr - part2);
-            shell_annunciators(-1, -1, 0, -1, -1, -1);
+            set_annunciators(-1, -1, 1, -1, -1, -1);
+            sbuf = NULL;
+            pbuf = buf;
+            slen = bufptr + vlength;
+            if (slen > 64) {
+                sbuf = (char *) malloc(slen);
+                if (sbuf == NULL) {
+                    slen = 64;
+                } else {
+                    memcpy(sbuf, buf, bufptr);
+                    pbuf = sbuf;
+                }
+            }
+            bufptr += vartype2string(v, pbuf + bufptr, slen - bufptr);
+            print_lines(pbuf, bufptr, true);
+            free(sbuf);
+            set_annunciators(-1, -1, 0, -1, -1, -1);
         } else
-            return ERR_STOP;
+            err = ERR_STOP;
     }
-    return ERR_NONE;
+    free_vartype(v);
+    return err;
 }
 
 int docmd_view(arg_struct *arg) {
@@ -504,7 +547,8 @@ int docmd_xeq(arg_struct *arg) {
             int4 dummy2;
             bool dummy3;
             pop_rtn_addr(&dummy1, &dummy2, &dummy3);
-        }
+        } else
+            save_csld();
         return err;
     } else {
         directory *dir = cwd;
@@ -512,6 +556,7 @@ int docmd_xeq(arg_struct *arg) {
         if (err != ERR_NONE)
             return err;
         clear_all_rtns();
+        save_csld();
         // GTO with ARGTYPE_LBLINDEX (which can only happen in RUN mode),
         // sets the current directory to the target's containing directory.
         // XEQ shouldn't do this, hence:
@@ -814,7 +859,8 @@ static void pixel_helper(phloat dx, phloat dy) {
     dy = dy < 0 ? -floor(-dy + 0.5) : floor(dy + 0.5);
     int w1 = disp_w + 1;
     int x = dx < -w1 ? -w1 : dx > w1 ? w1 : to_int(dx);
-    int y = dy < -w1 ? -w1 : dy > w1 ? w1 : to_int(dy);
+    int h1 = disp_h + 1;
+    int y = dy < -h1 ? -h1 : dy > h1 ? h1 : to_int(dy);
     int i;
     int dot = 1;
     if (x < 0) {
@@ -988,19 +1034,19 @@ struct sigma_label_spec {
 };
 
 static const sigma_label_spec sigma_labels[] = {
-    { "\005X=",           3 },
-    { "\005X\0362=",      5 },
-    { "\005Y=",           3 },
-    { "\005Y\0362=",      5 },
-    { "\005XY=",          4 },
-    { "N=",               2 },
-    { "\005LN(X)=",       7 },
-    { "\005LN(X)\0362=",  9 },
-    { "\005LN(Y)=",       7 },
-    { "\005LN(Y)\0362=",  9 },
-    { "\005LN(X)LN(Y)=", 12 },
-    { "\005XLN(Y)=",      8 },
-    { "\005YLN(X)=",      8 }
+    { "\5X=",           3 },
+    { "\5X^2=",         5 },
+    { "\5Y=",           3 },
+    { "\5Y^2=",         5 },
+    { "\5XY=",          4 },
+    { "N=",             2 },
+    { "\5LN(X)=",       7 },
+    { "\5LN(X)^2=",     9 },
+    { "\5LN(Y)=",       7 },
+    { "\5LN(Y)^2=",     9 },
+    { "\5LN(X)LN(Y)=", 12 },
+    { "\5XLN(Y)=",      8 },
+    { "\5YLN(X)=",      8 }
 };
 
 int docmd_prsigma(arg_struct *arg) {
@@ -1026,7 +1072,7 @@ int docmd_prsigma(arg_struct *arg) {
     if (!flags.f.printer_exists)
         return ERR_PRINTING_IS_DISABLED;
 
-    shell_annunciators(-1, -1, 1, -1, -1, -1);
+    set_annunciators(-1, -1, 1, -1, -1, -1);
     print_text(NULL, 0, true);
     for (i = 0; i < nr; i++) {
         int4 j = i + mode_sigma_reg;
@@ -1049,7 +1095,7 @@ int docmd_prsigma(arg_struct *arg) {
             print_wide(sigma_labels[i].text, sigma_labels[i].length, buf, bufptr);
         }
     }
-    shell_annunciators(-1, -1, 0, -1, -1, -1);
+    set_annunciators(-1, -1, 0, -1, -1, -1);
     return ERR_NONE;
 }
 
@@ -1075,6 +1121,7 @@ int docmd_prp(arg_struct *arg) {
 
 static vartype *prv_var;
 static int4 prv_index;
+static bool prv_prreg;
 static int prv_worker(bool interrupted);
 
 int docmd_prv(arg_struct *arg) {
@@ -1097,7 +1144,7 @@ int docmd_prv(arg_struct *arg) {
         if (!flags.f.printer_exists)
             return ERR_PRINTING_IS_DISABLED;
 
-        shell_annunciators(-1, -1, 1, -1, -1, -1);
+        set_annunciators(-1, -1, 1, -1, -1, -1);
         string2buf(lbuf, 8, &llen, arg->val.text, arg->length);
         char2buf(lbuf, 8, &llen, '=');
         if (v->type == TYPE_STRING || v->type == TYPE_EQUATION) {
@@ -1118,7 +1165,7 @@ int docmd_prv(arg_struct *arg) {
             }
             char *sbuf = (char *) malloc(length + 2);
             if (sbuf == NULL) {
-                shell_annunciators(-1, -1, 0, -1, -1, -1);
+                set_annunciators(-1, -1, 0, -1, -1, -1);
                 return ERR_INSUFFICIENT_MEMORY;
             }
             sbuf[0] = d;
@@ -1136,12 +1183,13 @@ int docmd_prv(arg_struct *arg) {
                 || v->type == TYPE_LIST
                 && ((vartype_list *) v)->size > 0) {
             prv_var = v;
+            prv_prreg = false;
             prv_index = 0;
             mode_interruptible = prv_worker;
             mode_stoppable = true;
             return ERR_INTERRUPTIBLE;
         } else {
-            shell_annunciators(-1, -1, 0, -1, -1, -1);
+            set_annunciators(-1, -1, 0, -1, -1, -1);
             return ERR_NONE;
         }
     }
@@ -1153,19 +1201,27 @@ static int prv_worker(bool interrupted) {
     int4 i, j, sz;
 
     if (interrupted) {
-        shell_annunciators(-1, -1, 0, -1, -1, -1);
+        set_annunciators(-1, -1, 0, -1, -1, -1);
         return ERR_STOP;
     }
 
     if (prv_var->type == TYPE_REALMATRIX) {
         vartype_realmatrix *rm = (vartype_realmatrix *) prv_var;
-        i = prv_index / rm->columns;
-        j = prv_index % rm->columns;
         sz = rm->rows * rm->columns;
-        llen = int2string(i + 1, lbuf, 32);
-        char2buf(lbuf, 32, &llen, ':');
-        llen += int2string(j + 1, lbuf + llen, 32 - llen);
-        char2buf(lbuf, 32, &llen, '=');
+        if (prv_prreg) {
+            char2buf(lbuf, 32, &llen, 'R');
+            if (prv_index < 10)
+                char2buf(lbuf, 32, &llen, '0');
+            llen += int2string(prv_index, lbuf + llen, 32 - llen);
+            char2buf(lbuf, 32, &llen, '=');
+        } else {
+            i = prv_index / rm->columns;
+            j = prv_index % rm->columns;
+            llen = int2string(i + 1, lbuf, 32);
+            char2buf(lbuf, 32, &llen, ':');
+            llen += int2string(j + 1, lbuf + llen, 32 - llen);
+            char2buf(lbuf, 32, &llen, '=');
+        }
         if (rm->array->is_string[prv_index] != 0) {
             char *text;
             int4 len;
@@ -1187,15 +1243,23 @@ static int prv_worker(bool interrupted) {
         }
     } else if (prv_var->type == TYPE_COMPLEXMATRIX) {
         vartype_complexmatrix *cm = (vartype_complexmatrix *) prv_var;
+        sz = cm->rows * cm->columns;
+        if (prv_prreg) {
+            char2buf(lbuf, 32, &llen, 'R');
+            if (prv_index < 10)
+                char2buf(lbuf, 32, &llen, '0');
+            llen += int2string(prv_index, lbuf + llen, 32 - llen);
+            char2buf(lbuf, 32, &llen, '=');
+        } else {
+            i = prv_index / cm->columns;
+            j = prv_index % cm->columns;
+            llen = int2string(i + 1, lbuf, 32);
+            char2buf(lbuf, 32, &llen, ':');
+            llen += int2string(j + 1, lbuf + llen, 32 - llen);
+            char2buf(lbuf, 32, &llen, '=');
+        }
         vartype_complex cpx;
         cpx.type = TYPE_COMPLEX;
-        i = prv_index / cm->columns;
-        j = prv_index % cm->columns;
-        sz = cm->rows * cm->columns;
-        llen = int2string(i + 1, lbuf, 32);
-        char2buf(lbuf, 32, &llen, ':');
-        llen += int2string(j + 1, lbuf + llen, 32 - llen);
-        char2buf(lbuf, 32, &llen, '=');
         cpx.re = cm->array->data[2 * prv_index];
         cpx.im = cm->array->data[2 * prv_index + 1];
         rlen = vartype2string((vartype *) &cpx, rbuf, 100);
@@ -1241,11 +1305,29 @@ static int prv_worker(bool interrupted) {
     if (++prv_index < sz)
         return ERR_INTERRUPTIBLE;
     else {
-        shell_annunciators(-1, -1, 0, -1, -1, -1);
+        set_annunciators(-1, -1, 0, -1, -1, -1);
         return ERR_NONE;
     }
 }
 
+int docmd_prreg(arg_struct *arg) {
+    vartype *regs = recall_var("REGS", 4);
+    if (regs == NULL)
+        return ERR_NONEXISTENT;
+    if (!flags.f.printer_enable && program_running())
+        return ERR_NONE;
+    if (!flags.f.printer_exists)
+        return ERR_PRINTING_IS_DISABLED;
+    set_annunciators(-1, -1, 1, -1, -1, -1);
+    print_text(NULL, 0, true);
+    prv_var = regs;
+    prv_prreg = true;
+    prv_index = 0;
+    mode_interruptible = prv_worker;
+    mode_stoppable = true;
+    return ERR_INTERRUPTIBLE;
+}
+
 int docmd_prstk(arg_struct *arg) {
     char buf[100];
     int len;
@@ -1254,7 +1336,7 @@ int docmd_prstk(arg_struct *arg) {
         return ERR_NONE;
     if (!flags.f.printer_exists)
         return ERR_PRINTING_IS_DISABLED;
-    shell_annunciators(-1, -1, 1, -1, -1, -1);
+    set_annunciators(-1, -1, 1, -1, -1, -1);
     if (arg != NULL)
         print_text(NULL, 0, true);
     if (flags.f.big_stack) {
@@ -1306,7 +1388,7 @@ int docmd_prstk(arg_struct *arg) {
             }
         }
     }
-    shell_annunciators(-1, -1, 0, -1, -1, -1);
+    set_annunciators(-1, -1, 0, -1, -1, -1);
     return ERR_NONE;
 }
 
@@ -1325,7 +1407,7 @@ int docmd_prx(arg_struct *arg) {
     if (!flags.f.printer_exists)
         return ERR_PRINTING_IS_DISABLED;
     else {
-        shell_annunciators(-1, -1, 1, -1, -1, -1);
+        set_annunciators(-1, -1, 1, -1, -1, -1);
         if (stack[sp]->type == TYPE_STRING || stack[sp]->type == TYPE_EQUATION) {
             const char *text;
             int length;
@@ -1344,7 +1426,7 @@ int docmd_prx(arg_struct *arg) {
             }
             char *lbuf = (char *) malloc(length + 2);
             if (lbuf == NULL) {
-                shell_annunciators(-1, -1, 0, -1, -1, -1);
+                set_annunciators(-1, -1, 0, -1, -1, -1);
                 return ERR_INSUFFICIENT_MEMORY;
             }
             lbuf[0] = d;
@@ -1361,7 +1443,7 @@ int docmd_prx(arg_struct *arg) {
             u->type = TYPE_UNIT;
             char *text = (char *) malloc(len + 1 + u->length);
             if (text == NULL) {
-                shell_annunciators(-1, -1, 0, -1, -1, -1);
+                set_annunciators(-1, -1, 0, -1, -1, -1);
                 return ERR_INSUFFICIENT_MEMORY;
             }
             memcpy(text, buf, len);
@@ -1391,12 +1473,13 @@ int docmd_prx(arg_struct *arg) {
                             || stack[sp]->type == TYPE_LIST
                             && ((vartype_list *) stack[sp])->size > 0)) {
             prv_var = stack[sp];
+            prv_prreg = false;
             prv_index = 0;
             mode_interruptible = prv_worker;
             mode_stoppable = true;
             return ERR_INTERRUPTIBLE;
         } else {
-            shell_annunciators(-1, -1, 0, -1, -1, -1);
+            set_annunciators(-1, -1, 0, -1, -1, -1);
             return ERR_NONE;
         }
     }
@@ -1412,7 +1495,7 @@ int docmd_prusr(arg_struct *arg) {
     if (!flags.f.printer_exists)
         return ERR_PRINTING_IS_DISABLED;
     else {
-        shell_annunciators(-1, -1, 1, -1, -1, -1);
+        set_annunciators(-1, -1, 1, -1, -1, -1);
         print_text(NULL, 0, true);
         prusr_state = arg == NULL ? 2 : 0;
         prusr_index = cwd->vars_count - 1;
@@ -1424,14 +1507,14 @@ int docmd_prusr(arg_struct *arg) {
 
 static int prusr_worker(bool interrupted) {
     if (interrupted) {
-        shell_annunciators(-1, -1, 0, -1, -1, -1);
+        set_annunciators(-1, -1, 0, -1, -1, -1);
         return ERR_STOP;
     }
 
     if (prusr_state != 1) {
         if (prusr_index < 0) {
             if (prusr_state == 2) {
-                shell_annunciators(-1, -1, 0, -1, -1, -1);
+                set_annunciators(-1, -1, 0, -1, -1, -1);
                 return ERR_NONE;
             } else {
                 if (cwd->vars_count > 0)
@@ -1470,7 +1553,7 @@ static int prusr_worker(bool interrupted) {
         state1:
         len = 0;
         if (prusr_index >= cwd->labels_count) {
-            shell_annunciators(-1, -1, 0, -1, -1, -1);
+            set_annunciators(-1, -1, 0, -1, -1, -1);
             return ERR_NONE;
         }
         if (cwd->labels[prusr_index].length == 0) {
@@ -1503,9 +1586,9 @@ int docmd_list(arg_struct *arg) {
 int docmd_adv(arg_struct *arg) {
     if (flags.f.printer_exists
             && (flags.f.printer_enable || !program_running())) {
-        shell_annunciators(-1, -1, 1, -1, -1, -1);
+        set_annunciators(-1, -1, 1, -1, -1, -1);
         print_text(NULL, 0, true);
-        shell_annunciators(-1, -1, 0, -1, -1, -1);
+        set_annunciators(-1, -1, 0, -1, -1, -1);
     }
     return ERR_NONE;
 }
@@ -1516,9 +1599,9 @@ int docmd_prlcd(arg_struct *arg) {
     if (!flags.f.printer_exists)
         return ERR_PRINTING_IS_DISABLED;
     else {
-        shell_annunciators(-1, -1, 1, -1, -1, -1);
+        set_annunciators(-1, -1, 1, -1, -1, -1);
         int err = print_display();
-        shell_annunciators(-1, -1, 0, -1, -1, -1);
+        set_annunciators(-1, -1, 0, -1, -1, -1);
         return err;
     }
 }
@@ -1708,25 +1791,13 @@ int docmd_stop(arg_struct *arg) {
 int docmd_newmat(arg_struct *arg) {
     vartype *m;
 
-    phloat x = ((vartype_real *) stack[sp])->x;
-    if (x <= -2147483648.0 || x >= 2147483648.0)
+    int4 row, col;
+    if (!dim_to_int4(stack[sp - 1], &row))
         return ERR_DIMENSION_ERROR;
-    int4 xx = to_int4(x);
-    if (xx == 0)
+    if (!dim_to_int4(stack[sp], &col))
         return ERR_DIMENSION_ERROR;
-    if (xx < 0)
-        xx = -xx;
 
-    phloat y = ((vartype_real *) stack[sp - 1])->x;
-    if (y <= -2147483648.0 || y >= 2147483648.0)
-        return ERR_DIMENSION_ERROR;
-    int4 yy = to_int4(y);
-    if (yy == 0)
-        return ERR_DIMENSION_ERROR;
-    if (yy < 0)
-        yy = -yy;
-
-    m = new_realmatrix(yy, xx);
+    m = new_realmatrix(row + 1, col + 1);
     if (m == NULL)
         return ERR_INSUFFICIENT_MEMORY;
     else
