@@ -1,6 +1,6 @@
 /*****************************************************************************
  * Plus42 -- an enhanced HP-42S calculator simulator
- * Copyright (C) 2004-2022  Thomas Okken
+ * Copyright (C) 2004-2025  Thomas Okken
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License, version 2,
@@ -17,6 +17,7 @@
 
 #include <stdlib.h>
 #include <string.h>
+#include <limits.h>
 
 #include "core_commands1.h"
 #include "core_commands2.h"
@@ -342,7 +343,7 @@ int docmd_custom(arg_struct *arg) {
     if (mode_plainmenu != MENU_CUSTOM1
             && mode_plainmenu != MENU_CUSTOM2
             && mode_plainmenu != MENU_CUSTOM3)
-        set_menu(MENULEVEL_PLAIN, MENU_CUSTOM1);
+        set_plainmenu(MENU_CUSTOM1, NULL, 0);
     return ERR_NONE;
 }
 
@@ -350,39 +351,33 @@ int docmd_delr(arg_struct *arg) {
     vartype *m, *newx;
     vartype_realmatrix *rm;
     vartype_complexmatrix *cm;
+    vartype_list *list;
     int4 rows, columns, i, j, n, newi;
     int err, refcount;
     int interactive;
 
-    switch (matedit_mode) {
-        case 0:
-            return ERR_NONEXISTENT;
-        case 1:
-        case 3:
-            m = matedit_get();
-            break;
-        case 2:
-            m = matedit_x;
-            break;
-        default:
-            return ERR_INTERNAL_ERROR;
-    }
-    if (m == NULL)
-        return ERR_NONEXISTENT;
+    err = matedit_get(&m);
+    if (err != ERR_NONE)
+        return err;
 
     if (m->type == TYPE_REALMATRIX) {
         rm = (vartype_realmatrix *) m;
         rows = rm->rows;
         columns = rm->columns;
         refcount = rm->array->refcount;
-    } else {
+    } else if (m->type == TYPE_COMPLEXMATRIX) {
         cm = (vartype_complexmatrix *) m;
         rows = cm->rows;
         columns = cm->columns;
         refcount = cm->array->refcount;
+    } else {
+        list = (vartype_list *) m;
+        rows = list->size;
+        columns = 1;
+        refcount = list->array->refcount;
     }
 
-    if (rows == 1)
+    if (rows == 1 && m->type != TYPE_LIST)
         return ERR_DIMENSION_ERROR;
 
     n = (matedit_i + 1) * columns + matedit_j;
@@ -409,9 +404,15 @@ int docmd_delr(arg_struct *arg) {
                 newx = new_string(text, len);
             } else
                 newx = new_real(rm->array->data[n]);
-        } else
+        } else if (m->type == TYPE_COMPLEXMATRIX) {
             newx = new_complex(cm->array->data[2 * n],
                                cm->array->data[2 * n + 1]);
+        } else {
+            if (n < 0)
+                newx = new_real(0);
+            else
+                newx = dup_vartype(list->array->data[n]);
+        }
 
         if (newx == NULL)
             return ERR_INSUFFICIENT_MEMORY;
@@ -457,7 +458,7 @@ int docmd_delr(arg_struct *arg) {
                     free_vartype(newx);
                 return err;
             }
-        } else {
+        } else if (m->type == TYPE_COMPLEXMATRIX) {
             for (j = 0; j < 2 * columns; j++) {
                 phloat tempd = cm->array->data[matedit_i * 2 * columns + j];
                 for (i = matedit_i; i < rows - 1; i++)
@@ -480,6 +481,13 @@ int docmd_delr(arg_struct *arg) {
                     free_vartype(newx);
                 return err;
             }
+        } else /* m->type == TYPE_LIST */ {
+            /* This one is easy because shrinking an unshared list always succeeds. */
+            vartype *tmp = list->array->data[matedit_i];
+            memmove(list->array->data + matedit_i, list->array->data + matedit_i + 1,
+                    (rows - matedit_i - 1) * sizeof(vartype *));
+            list->array->data[rows - 1] = tmp;
+            dimension_array_ref(m, rows - 1, 1);
         }
     } else {
         /* We're sharing this array. I don't use disentangle() because it
@@ -523,7 +531,7 @@ int docmd_delr(arg_struct *arg) {
             rm->array->refcount--;
             rm->array = array;
             rm->rows--;
-        } else {
+        } else if (m->type == TYPE_COMPLEXMATRIX) {
             complexmatrix_data *array = (complexmatrix_data *)
                                 malloc(sizeof(complexmatrix_data));
             if (array == NULL) {
@@ -546,6 +554,36 @@ int docmd_delr(arg_struct *arg) {
             cm->array->refcount--;
             cm->array = array;
             cm->rows--;
+        } else /* m->type == TYPE_LIST */ {
+            list_data *array = (list_data *) malloc(sizeof(list_data));
+            if (array == NULL) {
+                if (interactive)
+                    free_vartype(newx);
+                return ERR_INSUFFICIENT_MEMORY;
+            }
+            array->data = (vartype **) malloc(newsize * sizeof(vartype *));
+            if (array->data == NULL) {
+                if (interactive)
+                    free_vartype(newx);
+                free(array);
+                return ERR_INSUFFICIENT_MEMORY;
+            }
+            for (int4 i = 0; i < newsize; i++) {
+                array->data[i] = dup_vartype(list->array->data[i < matedit_i ? i : i + 1]);
+                if (array->data[i] == NULL) {
+                    for (int4 j = 0; j < i; j++)
+                        free_vartype(array->data[j]);
+                    if (interactive)
+                        free_vartype(newx);
+                    free(array->data);
+                    free(array);
+                    return ERR_INSUFFICIENT_MEMORY;
+                }
+            }
+            array->refcount = 1;
+            list->array->refcount--;
+            list->array = array;
+            list->size--;
         }
     }
     if (interactive) {
@@ -554,14 +592,14 @@ int docmd_delr(arg_struct *arg) {
         else
             free_vartype(stack[sp]);
         stack[sp] = newx;
+        mode_disable_stack_lift = true;
     }
     matedit_i = newi;
     return ERR_NONE;
 }
 
-static void det_completion(int error, vartype *det) {
-    if (error == ERR_NONE)
-        unary_result(det);
+static void det_completion(vartype *det) {
+    unary_result(det);
 }
 
 int docmd_det(arg_struct *arg) {
@@ -746,37 +784,29 @@ int docmd_dot(arg_struct *arg) {
     return binary_result(v);
 }
 
-int matedit_get_dim(int4 *rows, int4 *columns) {
+int matedit_get_dim(int4 *rows, int4 *columns, vartype **res) {
     vartype *m;
 
-    switch (matedit_mode) {
-        case 0:
-            return ERR_NONEXISTENT;
-        case 1:
-        case 3:
-            m = matedit_get();
-            break;
-        case 2:
-            m = matedit_x;
-            break;
-        default:
-            return ERR_INTERNAL_ERROR;
-    }
-
-    if (m == NULL)
-        return ERR_NONEXISTENT;
+    int err = matedit_get(&m);
+    if (err != ERR_NONE)
+        return err;
 
     if (m->type == TYPE_REALMATRIX) {
         vartype_realmatrix *rm = (vartype_realmatrix *) m;
         *rows = rm->rows;
         *columns = rm->columns;
-        return ERR_NONE;
-    } else {
+    } else if (m->type == TYPE_COMPLEXMATRIX) {
         vartype_complexmatrix *cm = (vartype_complexmatrix *) m;
         *rows = cm->rows;
         *columns = cm->columns;
-        return ERR_NONE;
+    } else { // TYPE_LIST
+        vartype_list *list = (vartype_list *) m;
+        *rows = list->size;
+        *columns = 1;
     }
+    if (res != NULL)
+        *res = m;
+    return ERR_NONE;
 }
 
 /* NOTE: this is a callback for set_menu(); it is declared in
@@ -797,6 +827,9 @@ int appmenu_exitcallback_1(int menuid, bool exitall) {
             err = ERR_NONE;
         else
             err = docmd_stoel(NULL);
+        if (err == ERR_NONEXISTENT)
+            /* This can happen when editing an empty list; ignore. */
+            err = ERR_NONE;
         if (err != ERR_NONE && err != ERR_INSUFFICIENT_MEMORY) {
             /* Reinstate the callback so we'll get called again when
              * the user tries to leave again.
@@ -816,6 +849,9 @@ int appmenu_exitcallback_1(int menuid, bool exitall) {
             matedit_x = NULL;
         }
         matedit_mode = 0;
+        free(matedit_stack);
+        matedit_stack = NULL;
+        matedit_stack_depth = 0;
         flags.f.grow = 0;
         flags.f.stack_lift_disable = 0;
         /* Note: no need to check the value returned by set_menu() here:
@@ -848,44 +884,55 @@ int docmd_edit(arg_struct *arg) {
     int err = finish_edit();
     if (err != ERR_NONE)
         return err;
-    if (stack[sp]->type == TYPE_REALMATRIX || stack[sp]->type == TYPE_COMPLEXMATRIX) {
-        vartype *v;
-        if (stack[sp]->type == TYPE_REALMATRIX) {
-            vartype_realmatrix *rm = (vartype_realmatrix *) stack[sp];
-            if (rm->array->is_string[0] != 0) {
-                char *text;
-                int4 len;
-                get_matrix_string(rm, 0, &text, &len);
-                v = new_string(text, len);
-            } else
-                v = new_real(rm->array->data[0]);
-        } else {
-            vartype_complexmatrix *cm = (vartype_complexmatrix *) stack[sp];
-            v = new_complex(cm->array->data[0], cm->array->data[1]);
-        }
-        if (v == NULL)
-            return ERR_INSUFFICIENT_MEMORY;
-        matedit_mode = 2;
-        flags.f.grow = 0;
-        matedit_x = stack[sp];
-        stack[sp] = v;
-        matedit_i = 0;
-        matedit_j = 0;
-        if (mode_appmenu >= MENU_MATRIX1 && mode_appmenu <= MENU_MATRIX_SIMQ)
-            matedit_prev_appmenu = mode_appmenu;
-        else
-            matedit_prev_appmenu = MENU_NONE;
-        set_menu(MENULEVEL_APP, MENU_MATRIX_EDIT1);
-        set_appmenu_exitcallback(1);
-        print_trace();
-        if (flags.f.big_stack)
-            mode_disable_stack_lift = true;
-        else
-            /* HP-42S bug compatibility */
-            mode_disable_stack_lift = flags.f.stack_lift_disable;
-        return ERR_NONE;
-    } else
+    if (stack[sp]->type != TYPE_REALMATRIX
+            && stack[sp]->type != TYPE_COMPLEXMATRIX
+            && stack[sp]->type != TYPE_LIST)
         return ERR_INVALID_TYPE;
+
+    vartype *v;
+    if (stack[sp]->type == TYPE_REALMATRIX) {
+        vartype_realmatrix *rm = (vartype_realmatrix *) stack[sp];
+        if (rm->array->is_string[0] != 0) {
+            char *text;
+            int4 len;
+            get_matrix_string(rm, 0, &text, &len);
+            v = new_string(text, len);
+        } else
+            v = new_real(rm->array->data[0]);
+    } else if (stack[sp]->type == TYPE_COMPLEXMATRIX) {
+        vartype_complexmatrix *cm = (vartype_complexmatrix *) stack[sp];
+        v = new_complex(cm->array->data[0], cm->array->data[1]);
+    } else {
+        vartype_list *list = (vartype_list *) stack[sp];
+        if (list->size == 0)
+            v = new_real(0);
+        else
+            v = dup_vartype(list->array->data[0]);
+    }
+    if (v == NULL)
+        return ERR_INSUFFICIENT_MEMORY;
+    matedit_mode = 2;
+    flags.f.grow = 0;
+    matedit_x = stack[sp];
+    stack[sp] = v;
+    matedit_i = 0;
+    matedit_j = 0;
+    matedit_is_list = matedit_x->type == TYPE_LIST;
+    flags.f.matrix_edge_wrap = 0;
+    flags.f.matrix_end_wrap = 0;
+    if (mode_appmenu >= MENU_MATRIX1 && mode_appmenu <= MENU_MATRIX_SIMQ)
+        matedit_prev_appmenu = mode_appmenu;
+    else
+        matedit_prev_appmenu = MENU_NONE;
+    set_menu(MENULEVEL_APP, MENU_MATRIX_EDIT1);
+    set_appmenu_exitcallback(1);
+    print_trace();
+    if (flags.f.big_stack)
+        mode_disable_stack_lift = true;
+    else
+        /* HP-42S bug compatibility */
+        mode_disable_stack_lift = flags.f.stack_lift_disable;
+    return ERR_NONE;
 }
 
 int docmd_editn(arg_struct *arg) {
@@ -912,7 +959,9 @@ int docmd_editn(arg_struct *arg) {
     if (!mi.writable())
         return ERR_VARIABLE_NOT_WRITABLE;
     m = mi.value();
-    if (m->type != TYPE_REALMATRIX && m->type != TYPE_COMPLEXMATRIX)
+    if (m->type != TYPE_REALMATRIX
+            && m->type != TYPE_COMPLEXMATRIX
+            && m->type != TYPE_LIST)
         return ERR_INVALID_TYPE;
 
     vartype *v;
@@ -925,9 +974,15 @@ int docmd_editn(arg_struct *arg) {
             v = new_string(text, len);
         } else
             v = new_real(rm->array->data[0]);
-    } else {
+    } else if (m->type == TYPE_COMPLEXMATRIX) {
         vartype_complexmatrix *cm = (vartype_complexmatrix *) m;
         v = new_complex(cm->array->data[0], cm->array->data[1]);
+    } else {
+        vartype_list *list = (vartype_list *) m;
+        if (list->size == 0)
+            v = new_real(0);
+        else
+            v = dup_vartype(list->array->data[0]);
     }
     if (v == NULL)
         return ERR_INSUFFICIENT_MEMORY;
@@ -951,6 +1006,9 @@ int docmd_editn(arg_struct *arg) {
     stack[sp] = v;
     matedit_i = 0;
     matedit_j = 0;
+    matedit_is_list = m->type == TYPE_LIST;
+    flags.f.matrix_edge_wrap = 0;
+    flags.f.matrix_end_wrap = 0;
     if (mode_appmenu >= MENU_MATRIX1 && mode_appmenu <= MENU_MATRIX_SIMQ)
         matedit_prev_appmenu = mode_appmenu;
     else
@@ -970,7 +1028,7 @@ int docmd_exitall(arg_struct *arg) {
     return set_menu_return_err(MENULEVEL_APP, MENU_NONE, true);
 }
 
-static int mappable_e_pow_x_1(phloat x, phloat *y) {
+static int mappable_e_pow_x_1_r(phloat x, phloat *y) {
     *y = expm1(x);
     if (p_isinf(*y) != 0) {
         if (flags.f.range_error_ignore)
@@ -981,55 +1039,78 @@ static int mappable_e_pow_x_1(phloat x, phloat *y) {
     return ERR_NONE;
 }
 
+static int mappable_e_pow_x_1_c(phloat xre, phloat xim, phloat *yre, phloat *yim) {
+    phloat k = expm1(xre);
+    phloat t = xim;
+    phloat s = sin(t);
+    t = sin(t / 2);
+    t = -2 * t * t;
+    s *= k + 1;
+    if (-t > fabs(k))
+        *yre = k + k * t + t;
+    else
+        *yre = t + k * t + k;
+    *yim = s;
+    if (p_isinf(*yre) || p_isinf(*yim)) {
+        if (flags.f.range_error_ignore) {
+            *yre = p_isnan(*yre) ? 0 : *yre < 0 ? NEG_HUGE_PHLOAT : POS_HUGE_PHLOAT;
+            *yim = p_isnan(*yim) ? 0 : *yim < 0 ? NEG_HUGE_PHLOAT : POS_HUGE_PHLOAT;
+        } else
+            return ERR_OUT_OF_RANGE;
+    }
+    return ERR_NONE;
+}
+
 int docmd_e_pow_x_1(arg_struct *arg) {
     vartype *v;
-    int err = map_unary(stack[sp], &v, mappable_e_pow_x_1, NULL);
+    int err = map_unary(stack[sp], &v, mappable_e_pow_x_1_r, NULL);
+    if (err == ERR_NONE)
+        unary_result(v);
+    return err;
+}
+
+int docmd_c_e_pow_x_1(arg_struct *arg) {
+    vartype *v;
+    int err = map_unary(stack[sp], &v, mappable_e_pow_x_1_r, mappable_e_pow_x_1_c);
     if (err == ERR_NONE)
         unary_result(v);
     return err;
 }
 
 static int fnrm(vartype *m, phloat *norm) {
+    int4 size;
+    phloat *data;
     if (m->type == TYPE_REALMATRIX) {
         vartype_realmatrix *rm = (vartype_realmatrix *) m;
         if (contains_strings(rm))
             return ERR_ALPHA_DATA_IS_INVALID;
-        int4 size = rm->rows * rm->columns;
-        phloat nrm = 0;
-        for (int4 i = 0; i < size; i++) {
-            /* TODO -- overflows in intermediaries */
-            phloat x = rm->array->data[i];
-            nrm += x * x;
-        }
-        if (p_isinf(nrm)) {
-            if (flags.f.range_error_ignore)
-                nrm = POS_HUGE_PHLOAT;
-            else
-                return ERR_OUT_OF_RANGE;
-        } else
-            nrm = sqrt(nrm);
-        *norm = nrm;
-        return ERR_NONE;
+        size = rm->rows * rm->columns;
+        data = rm->array->data;
     } else {
         vartype_complexmatrix *cm = (vartype_complexmatrix *) m;
-        int4 size = 2 * cm->rows * cm->columns;
-        int4 i;
-        phloat nrm = 0;
-        for (i = 0; i < size; i++) {
-            /* TODO -- overflows in intermediaries */
-            phloat x = cm->array->data[i];
-            nrm += x * x;
-        }
-        if (p_isinf(nrm)) {
-            if (flags.f.range_error_ignore)
-                nrm = POS_HUGE_PHLOAT;
-            else
-                return ERR_OUT_OF_RANGE;
-        } else
-            nrm = sqrt(nrm);
-        *norm = nrm;
-        return ERR_NONE;
+        size = 2 * cm->rows * cm->columns;
+        data = cm->array->data;
     }
+    int max_exp = INT_MIN;
+    for (int4 i = 0; i < size; i++) {
+        int s = ilogb(data[i]);
+        if (s > max_exp)
+            max_exp = s;
+    }
+    phloat nrm = 0;
+    for (int4 i = 0; i < size; i++) {
+        phloat x = scalbn(data[i], -max_exp);
+        nrm += x * x;
+    }
+    nrm = scalbn(sqrt(nrm), max_exp);
+    if (p_isinf(nrm)) {
+        if (flags.f.range_error_ignore)
+            nrm = POS_HUGE_PHLOAT;
+        else
+            return ERR_OUT_OF_RANGE;
+    }
+    *norm = nrm;
+    return ERR_NONE;
 }
 
 int docmd_fnrm(arg_struct *arg) {
@@ -1047,25 +1128,13 @@ int docmd_fnrm(arg_struct *arg) {
 
 int docmd_getm(arg_struct *arg) {
     vartype *m;
-    phloat xx, yy;
     int4 x, y;
 
-    switch (matedit_mode) {
-        case 0:
-            return ERR_NONEXISTENT;
-        case 1:
-        case 3:
-            m = matedit_get();
-            break;
-        case 2:
-            m = matedit_x;
-            break;
-        default:
-            return ERR_INTERNAL_ERROR;
-    }
-
-    if (m == NULL)
-        return ERR_NONEXISTENT;
+    int err = matedit_get(&m);
+    if (err != ERR_NONE)
+        return err;
+    if (m->type == TYPE_LIST)
+        return ERR_INVALID_TYPE;
 
     if (stack[sp]->type == TYPE_STRING)
         return ERR_ALPHA_DATA_IS_INVALID;
@@ -1076,23 +1145,12 @@ int docmd_getm(arg_struct *arg) {
     if (stack[sp - 1]->type != TYPE_REAL)
         return ERR_INVALID_TYPE;
 
-    xx = ((vartype_real *) stack[sp])->x;
-    if (xx <= -2147483648.0 || xx >= 2147483648.0)
+    if (!dim_to_int4(stack[sp], &x))
         return ERR_DIMENSION_ERROR;
-    x = to_int4(xx);
-    if (x == 0)
+    x++;
+    if (!dim_to_int4(stack[sp - 1], &y))
         return ERR_DIMENSION_ERROR;
-    if (x < 0)
-        x = -x;
-
-    yy = ((vartype_real *) stack[sp - 1])->x;
-    if (yy <= -2147483648.0 || yy >= 2147483648.0)
-        return ERR_DIMENSION_ERROR;
-    y = to_int4(yy);
-    if (y == 0)
-        return ERR_DIMENSION_ERROR;
-    if (y < 0)
-        y = -y;
+    y++;
 
     if (m->type == TYPE_REALMATRIX) {
         vartype_realmatrix *src, *dst;
@@ -1177,32 +1235,36 @@ static int hms_add_or_sub(bool add) {
 
         if (add) {
             ys += xs;
-            if (ys >= 60) {
+            while (ys >= 60) {
                 ys -= 60;
                 ym++;
             }
             ym += xm;
-            if (ym >= 60) {
+            while (ym >= 60) {
                 ym -= 60;
                 yh++;
             }
             yh += xh;
         } else {
             ys -= xs;
-            if (ys < 0) {
-                ys += 60;
-                ym--;
-            } else if (ys >= 60) {
+            if (ys >= 60) {
                 ys -= 60;
                 ym++;
+            } else {
+                while (ys < 0) {
+                    ys += 60;
+                    ym--;
+                }
             }
             ym -= xm;
-            if (ym < 0) {
-                ym += 60;
-                yh--;
-            } else if (ym >= 60) {
+            if (ym >= 60) {
                 ym -= 60;
                 yh++;
+            } else {
+                while (ym < 0) {
+                    ym += 60;
+                    yh--;
+                }
             }
             yh -= xh;
             if (yh < 0) {
@@ -1296,6 +1358,9 @@ int docmd_i_add(arg_struct *arg) {
     int err = matedit_get_dim(&rows, &columns);
     if (err != ERR_NONE)
         return err;
+    if (rows == 0)
+        // Empty list. We allow matedit_i == 0, because we have to allow *something*.
+        rows = 1;
     if (++matedit_i >= rows) {
         flags.f.matrix_edge_wrap = 1;
         matedit_i = 0;
@@ -1316,6 +1381,9 @@ int docmd_i_sub(arg_struct *arg) {
     int err = matedit_get_dim(&rows, &columns);
     if (err != ERR_NONE)
         return err;
+    if (rows == 0)
+        // Empty list. We allow matedit_i == 0, because we have to allow *something*.
+        rows = 1;
     if (--matedit_i < 0) {
         flags.f.matrix_edge_wrap = 1;
         matedit_i = rows - 1;
@@ -1335,6 +1403,9 @@ void matedit_goto(int4 row, int4 column) {
     int4 rows, columns;
     int err = matedit_get_dim(&rows, &columns);
     if (err == ERR_NONE) {
+        if (rows == 0)
+            // Empty list. We allow matedit_i == 0, because we have to allow *something*.
+            rows = 1;
         if (row == 0 || row > rows || column == 0 || column > columns)
             err = ERR_DIMENSION_ERROR;
         else {
@@ -1347,7 +1418,7 @@ void matedit_goto(int4 row, int4 column) {
         }
     }
     if (err != ERR_NONE) {
-        display_error(err, false);
+        display_error(err);
         flush_display();
     }
 }
@@ -1375,7 +1446,9 @@ int docmd_index(arg_struct *arg) {
     if (!mi.writable())
         return ERR_VARIABLE_NOT_WRITABLE;
     m = mi.value();
-    if (m->type != TYPE_REALMATRIX && m->type != TYPE_COMPLEXMATRIX)
+    if (m->type != TYPE_REALMATRIX
+            && m->type != TYPE_COMPLEXMATRIX
+            && m->type != TYPE_LIST)
         return ERR_INVALID_TYPE;
 
     /* If the matrix we're about to index is local, and another matrix
@@ -1389,6 +1462,9 @@ int docmd_index(arg_struct *arg) {
     matedit_mode = 1;
     matedit_i = 0;
     matedit_j = 0;
+    matedit_is_list = m->type == TYPE_LIST;
+    flags.f.matrix_edge_wrap = 0;
+    flags.f.matrix_end_wrap = 0;
     string_copy(matedit_name, &matedit_length, arg->val.text, arg->length);
     matedit_dir = mi.dir;
     return ERR_NONE;
