@@ -1,6 +1,6 @@
 /*****************************************************************************
  * Plus42 -- an enhanced HP-42S calculator simulator
- * Copyright (C) 2004-2022  Thomas Okken
+ * Copyright (C) 2004-2025  Thomas Okken
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License, version 2,
@@ -24,7 +24,7 @@ int linalg_div(const vartype *left, const vartype *right,
                              int (*completion)(int, vartype *));
 int linalg_mul(const vartype *left, const vartype *right,
                              int (*completion)(int, vartype *));
-int linalg_inv(const vartype *src, void (*completion)(int, vartype *));
-int linalg_det(const vartype *src, void (*completion)(int, vartype *));
+int linalg_inv(const vartype *src, void (*completion)(vartype *));
+int linalg_det(const vartype *src, void (*completion)(vartype *));
 
 #endif
