@@ -1,6 +1,6 @@
 /*****************************************************************************
  * Plus42 -- an enhanced HP-42S calculator simulator
- * Copyright (C) 2004-2022  Thomas Okken
+ * Copyright (C) 2004-2025  Thomas Okken
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License, version 2,
@@ -55,6 +55,13 @@ struct caller_info {
             return err != ERR_NONE ? err : keep_running ? ERR_NONE : ERR_STOP;
         }
     }
+    void equation_deleted(int equation_index) {
+        if (prev_prgm.dir == eq_dir->id && prev_prgm.idx == equation_index) {
+            prev_prgm.dir = cwd->id;
+            prev_prgm.idx = cwd->prgms_count - 1;
+            prev_pc = cwd->prgms[cwd->prgms_count - 1].size - 2;
+        }
+    }
 };
 
 /* Solver */
@@ -373,6 +380,11 @@ void reset_math() {
     reset_integ();
 }
 
+void math_equation_deleted(int eqn_index) {
+    solve.caller.equation_deleted(eqn_index);
+    integ.caller.equation_deleted(eqn_index);
+}
+
 void clean_stack(int prev_sp) {
     if (flags.f.big_stack && prev_sp != -2 && sp > prev_sp) {
         int excess = sp - prev_sp;
@@ -818,7 +830,7 @@ static int finish_solve(int message) {
     string_copy(arg.val.text, &dummy, solve.var_name, solve.var_length);
     arg.length = solve.var_length;
 
-    print = flags.f.trace_print && flags.f.printer_exists;
+    print = (flags.f.trace_print || flags.f.normal_print) && flags.f.printer_exists;
 
     if (!solve.caller.keep_running) {
         view_helper(&arg, print);
@@ -961,6 +973,9 @@ int return_to_solve(bool failure, bool stop) {
                 solve.best_x = solve.curr_x;
             }
         } else if (stack[sp]->type == TYPE_UNIT) {
+            vartype_unit *u = (vartype_unit *) stack[sp];
+            if (u->length == 2 && u->text[0] == 19 && (u->text[1] == 'C' || u->text[1] == 'F'))
+                goto real_result;
             bool saved_norm = flags.f.normal_print;
             bool saved_trace = flags.f.trace_print;
             flags.f.normal_print = false;
@@ -1629,8 +1644,8 @@ static int finish_integ() {
         normalize_unit(pu + "*" + ru, &ru);
         if (ru == "")
             goto real_result;
-        x = new_unit(integ.sum * integ.b * 0.75, ru.c_str(), ru.length());
-        y = new_unit(integ.eps, ru.c_str(), ru.length());
+        x = new_unit(integ.sum * integ.b * 0.75, ru.c_str(), (int) ru.length());
+        y = new_unit(integ.eps, ru.c_str(), (int) ru.length());
     }
     if (x == NULL || y == NULL) {
         free_vartype(x);
@@ -1655,10 +1670,10 @@ static int finish_integ() {
         char *buf = (char *) malloc(disp_c);
         freer f(buf);
         int bufptr = 0;
-        string2buf(buf, disp_c, &bufptr, "\003=", 2);
+        string2buf(buf, disp_c, &bufptr, "\3=", 2);
         bufptr += vartype2string(x, buf + bufptr, disp_c - bufptr);
         draw_message(0, buf, bufptr);
-        if (flags.f.trace_print && flags.f.printer_exists)
+        if ((flags.f.trace_print || flags.f.normal_print) && flags.f.printer_exists)
             print_wide(buf, 2, buf + 2, bufptr - 2);
     }
 
