@@ -1,6 +1,6 @@
 /*****************************************************************************
  * Plus42 -- an enhanced HP-42S calculator simulator
- * Copyright (C) 2004-2022  Thomas Okken
+ * Copyright (C) 2004-2025  Thomas Okken
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License, version 2,
@@ -356,276 +356,323 @@
 #define CMD_HEADING     317
 /* HP-41 Time Module & CX Time */
 #define CMD_ADATE       318
-#define CMD_ALMCAT      319
-#define CMD_ALMNOW      320
-#define CMD_ATIME       321
-#define CMD_ATIME24     322
-#define CMD_CLK12       323
-#define CMD_CLK24       324
-#define CMD_CLKT        325
-#define CMD_CLKTD       326
-#define CMD_CLOCK       327
-#define CMD_CORRECT     328
-#define CMD_DATE        329
-#define CMD_DATE_PLUS   330
-#define CMD_DDAYS       331
-#define CMD_DMY         332
-#define CMD_DOW         333
-#define CMD_MDY         334
-#define CMD_RCLAF       335
-#define CMD_RCLSW       336
-#define CMD_RUNSW       337
-#define CMD_SETAF       338
-#define CMD_SETDATE     339
-#define CMD_SETIME      340
-#define CMD_SETSW       341
-#define CMD_STOPSW      342
-#define CMD_SW          343
-#define CMD_T_PLUS_X    344
-#define CMD_TIME        345
-#define CMD_XYZALM      346
-#define CMD_CLALMA      347
-#define CMD_CLALMX      348
-#define CMD_CLRALMS     349
-#define CMD_RCLALM      350
-#define CMD_SWPT        351
+#define CMD_ATIME       319
+#define CMD_ATIME24     320
+#define CMD_CLK12       321
+#define CMD_CLK24       322
+#define CMD_DATE        323
+#define CMD_DATE_PLUS   324
+#define CMD_DDAYS       325
+#define CMD_DMY         326
+#define CMD_DOW         327
+#define CMD_MDY         328
+#define CMD_TIME        329
 /* Intel Decimal Floating-Point Math Library: self-test */
-#define CMD_FPTEST      352
+#define CMD_FPTEST      330
 /* Programming */
-#define CMD_LSTO        353
-#define CMD_SST_UP      354
-#define CMD_SST_RT      355
-#define CMD_WSIZE       356
-#define CMD_WSIZE_T     357
-#define CMD_YMD         358
-#define CMD_BSIGNED     359
-#define CMD_BWRAP       360
-#define CMD_BRESET      361
-#define CMD_GETKEY1     362
-#define CMD_LASTO       363
-#define CMD_LCLV        364
+#define CMD_LSTO        331
+#define CMD_SST_UP      332
+#define CMD_SST_RT      333
+#define CMD_YMD         334
+#define CMD_GETKEY1     335
+#define CMD_GETKEYA     336
+#define CMD_LASTO       337
+#define CMD_LCLV        338
+/* BASE enhancements */
+#define CMD_SL          339
+#define CMD_SR          340
+#define CMD_ASR         341
+#define CMD_SLN         342
+#define CMD_SRN         343
+#define CMD_ASRN        344
+#define CMD_LJ          345
+#define CMD_RJ          346
+#define CMD_RL          347
+#define CMD_RR          348
+#define CMD_RLC         349
+#define CMD_RRC         350
+#define CMD_RLN         351
+#define CMD_RRN         352
+#define CMD_RLCN        353
+#define CMD_RRCN        354
+#define CMD_SB          355
+#define CMD_CB          356
+#define CMD_B_T         357
+#define CMD_NUM_B       358
+#define CMD_MASKL       359
+#define CMD_MASKR       360
+#define CMD_SC          361
+#define CMD_CC          362
+#define CMD_C_T         363
+#define CMD_N_TO_BS     364
+#define CMD_N_TO_BD     365
+#define CMD_N_TO_BQ     366
+#define CMD_N_TO_DS     367
+#define CMD_N_TO_DD     368
+#define CMD_N_TO_DQ     369
+#define CMD_BS_TO_N     370
+#define CMD_BD_TO_N     371
+#define CMD_BQ_TO_N     372
+#define CMD_DS_TO_N     373
+#define CMD_DD_TO_N     374
+#define CMD_DQ_TO_N     375
+#define CMD_WSIZE       376
+#define CMD_WSIZE_T     377
+#define CMD_BSIGNED     378
+#define CMD_BWRAP       379
+#define CMD_BRESET      380
+#define CMD_DECINT      381
+#define CMD_BINSEP      382
+#define CMD_OCTSEP      383
+#define CMD_DECSEP      384
+#define CMD_HEXSEP      385
 /* Useful X-Fcn functions missing from the 42S */
-#define CMD_ANUM        365
-#define CMD_X_SWAP_F    366
-#define CMD_RCLFLAG     367
-#define CMD_STOFLAG     368
+#define CMD_ANUM        386
+#define CMD_X_SWAP_F    387
+#define CMD_RCLFLAG     388
+#define CMD_STOFLAG     389
 /* User-defined functions */
-#define CMD_FUNC        369
-#define CMD_ERRMSG      370
-#define CMD_ERRNO       371
-#define CMD_RTNYES      372
-#define CMD_RTNNO       373
-#define CMD_RTNERR      374
-#define CMD_STRACE      375
+#define CMD_FUNC        390
+#define CMD_ERRMSG      391
+#define CMD_ERRNO       392
+#define CMD_RTNYES      393
+#define CMD_RTNNO       394
+#define CMD_RTNERR      395
+#define CMD_STRACE      396
 /* Big Stack */
-#define CMD_4STK        376
-#define CMD_L4STK       377
-#define CMD_NSTK        378
-#define CMD_LNSTK       379
-#define CMD_DEPTH       380
-#define CMD_DROP        381
-#define CMD_DROPN       382
-#define CMD_DUP         383
-#define CMD_DUPN        384
-#define CMD_PICK        385
-#define CMD_UNPICK      386
-#define CMD_RDNN        387
-#define CMD_RUPN        388
+#define CMD_4STK        397
+#define CMD_L4STK       398
+#define CMD_NSTK        399
+#define CMD_LNSTK       400
+#define CMD_DEPTH       401
+#define CMD_DROP        402
+#define CMD_DROPN       403
+#define CMD_DUP         404
+#define CMD_DUPN        405
+#define CMD_PICK        406
+#define CMD_UNPICK      407
+#define CMD_RDNN        408
+#define CMD_RUPN        409
 /* Miscellaneous */
-#define CMD_NOP         389
-#define CMD_FMA         390
-#define CMD_PGMMENU     391
-#define CMD_PMEXEC      392
-#define CMD_PGMVAR      393
-#define CMD_VARMNU1     394
-#define CMD_X2LINE      395
-#define CMD_A2LINE      396
-#define CMD_A2PLINE     397
-#define CMD_RCOMPLX     398
-#define CMD_PCOMPLX     399
-#define CMD_SKIP        400
-#define CMD_CPXMAT_T    401
-#define CMD_TYPE_T      402
-/* (Skipping 403 because of single-byte equality checks with CMD_END) */
-#define CMD_A_THRU_F_2  404
-#define CMD_DROP_CANCL  405
+#define CMD_NOP         410
+#define CMD_FMA         411
+#define CMD_PGMMENU     412
+#define CMD_PMEXEC      413
+#define CMD_PGMVAR      414
+#define CMD_VARMNU1     415
+#define CMD_X2LINE      416
+#define CMD_A2LINE      417
+#define CMD_A2PLINE     418
+#define CMD_RCOMPLX     419
+#define CMD_PCOMPLX     420
+#define CMD_CAPS        421
+#define CMD_MIXED       422
+#define CMD_STATIC      423
+#define CMD_DYNAMIC     424
+#define CMD_SKIP        425
+#define CMD_CPXMAT_T    426
+#define CMD_TYPE_T      427
+#define CMD_A_THRU_F_2  428
+#define CMD_DROP_CANCL  429
+#define CMD_PRREG       430
+#define CMD_CSLD_T      431
+#define CMD_C_LN_1_X    432
+#define CMD_C_E_POW_X_1 433
+#define CMD_GETMI       434
+#define CMD_PUTMI       435
+#define CMD_GETLI       436
+#define CMD_PUTLI       437
+#define CMD_LOCK        438
+#define CMD_UNLOCK      439
+#define CMD_IDENT       440
 /* String & List Functions */
-#define CMD_XSTR        406
-#define CMD_XASTO       407
-#define CMD_LXASTO      408
-#define CMD_XVIEW       409
-#define CMD_APPEND      410
-#define CMD_EXTEND      411
-#define CMD_SUBSTR      412
-#define CMD_LENGTH      413
-#define CMD_HEAD        414
-#define CMD_REV         415
-#define CMD_POS         416
-#define CMD_S_TO_N      417
-#define CMD_N_TO_S      418
-#define CMD_C_TO_N      419
-#define CMD_N_TO_C      420
-#define CMD_LIST_T      421
-#define CMD_NEWLIST     422
-#define CMD_NEWSTR      423
-#define CMD_TO_LIST     424
-#define CMD_FROM_LIST   425
+#define CMD_XSTR        441
+#define CMD_XASTO       442
+#define CMD_LXASTO      443
+#define CMD_XVIEW       444
+#define CMD_APPEND      445
+#define CMD_EXTEND      446
+#define CMD_SUBSTR      447
+#define CMD_LENGTH      448
+#define CMD_HEAD        449
+#define CMD_REV         450
+#define CMD_POS         451
+#define CMD_S_TO_N      452
+#define CMD_N_TO_S      453
+#define CMD_NN_TO_S     454
+#define CMD_C_TO_N      455
+#define CMD_N_TO_C      456
+#define CMD_LIST_T      457
+#define CMD_NEWLIST     458
+#define CMD_TO_LIST     459
+#define CMD_FROM_LIST   460
 /* Generalized Comparisons */
-#define CMD_X_EQ_NN     426
-#define CMD_X_NE_NN     427
-#define CMD_X_LT_NN     428
-#define CMD_X_GT_NN     429
-#define CMD_X_LE_NN     430
-#define CMD_X_GE_NN     431
-#define CMD_0_EQ_NN     432
-#define CMD_0_NE_NN     433
-#define CMD_0_LT_NN     434
-#define CMD_0_GT_NN     435
-#define CMD_0_LE_NN     436
-#define CMD_0_GE_NN     437
+#define CMD_X_EQ_NN     461
+#define CMD_X_NE_NN     462
+#define CMD_X_LT_NN     463
+#define CMD_X_GT_NN     464
+#define CMD_X_LE_NN     465
+#define CMD_X_GE_NN     466
+#define CMD_0_EQ_NN     467
+#define CMD_0_NE_NN     468
+#define CMD_0_LT_NN     469
+#define CMD_0_GT_NN     470
+#define CMD_0_LE_NN     471
+#define CMD_0_GE_NN     472
 /* Statitics Sums */
-#define CMD_SN          438
-#define CMD_SX          439
-#define CMD_SX2         440
-#define CMD_SY          441
-#define CMD_SY2         442
-#define CMD_SXY         443
-#define CMD_SLNX        444
-#define CMD_SLNX2       445
-#define CMD_SLNY        446
-#define CMD_SLNY2       447
-#define CMD_SLNXLNY     448
-#define CMD_SXLNY       449
-#define CMD_SYLNX       450
+#define CMD_SN          473
+#define CMD_SX          474
+#define CMD_SX2         475
+#define CMD_SY          476
+#define CMD_SY2         477
+#define CMD_SXY         478
+#define CMD_SLNX        479
+#define CMD_SLNX2       480
+#define CMD_SLNY        481
+#define CMD_SLNY2       482
+#define CMD_SLNXLNY     483
+#define CMD_SXLNY       484
+#define CMD_SYLNX       485
 /* Equations */
-#define CMD_PARSE       451
-#define CMD_UNPARSE     452
-#define CMD_EVAL        453
-#define CMD_EQN_T       454
-#define CMD_STD         455
-#define CMD_COMP        456
-#define CMD_GTOL        457
-#define CMD_XEQL        458
-#define CMD_GSTO        459
-#define CMD_GRCL        460
-#define CMD_SVAR        461
-#define CMD_GETITEM     462
-#define CMD_GEN_EQ      463
-#define CMD_GEN_NE      464
-#define CMD_GEN_LT      465
-#define CMD_GEN_GT      466
-#define CMD_GEN_LE      467
-#define CMD_GEN_GE      468
-#define CMD_GEN_AND     469
-#define CMD_GEN_OR      470
-#define CMD_GEN_XOR     471
-#define CMD_GEN_NOT     472
-#define CMD_IF_T        473
-#define CMD_TRUNC       474
-#define CMD_DDAYSC      475
-#define CMD_GETEQN      476
-#define CMD_TO_PAR      477
-#define CMD_FDEPTH      478
-#define CMD_PUTITEM     479
-#define CMD_EVALN       480
-#define CMD_EVALNi      481
-#define CMD_EQNSLV      482
-#define CMD_EQNINT      483
-#define CMD_EQNSLVi     484
-#define CMD_EQNINTi     485
-#define CMD_EQNVAR      486
-#define CMD_EQNMENU     487
-#define CMD_EQNMNU1     488
-#define CMD_SPPV        489
-#define CMD_SPFV        490
-#define CMD_USPV        491
-#define CMD_USFV        492
-#define CMD_GEN_N       493
-#define CMD_GEN_I       494
-#define CMD_GEN_PV      495
-#define CMD_GEN_PMT     496
-#define CMD_GEN_FV      497
-#define CMD_RAISE       498
-#define CMD_DIRECT      499
-#define CMD_NUMERIC     500
+#define CMD_PARSE       486
+#define CMD_UNPARSE     487
+#define CMD_EVAL        488
+#define CMD_EQN_T       489
+#define CMD_STD         490
+#define CMD_COMP        491
+#define CMD_GTOL        492
+#define CMD_XEQL        493
+#define CMD_GSTO        494
+#define CMD_GRCL        495
+#define CMD_SVAR        496
+#define CMD_GETITEM     497
+#define CMD_GEN_EQ      498
+#define CMD_GEN_NE      499
+#define CMD_GEN_LT      500
+#define CMD_GEN_GT      501
+#define CMD_GEN_LE      502
+#define CMD_GEN_GE      503
+#define CMD_GEN_AND     504
+#define CMD_GEN_OR      505
+#define CMD_GEN_XOR     506
+#define CMD_GEN_NOT     507
+#define CMD_IF_T        508
+#define CMD_TRUNC       509
+#define CMD_DDAYSC      510
+#define CMD_GETEQN      511
+#define CMD_TO_PAR      512
+#define CMD_FSTART      513
+#define CMD_FSTACK      514
+#define CMD_PUTITEM     515
+#define CMD_EVALN       516
+#define CMD_EVALNi      517
+#define CMD_EQNSLV      518
+#define CMD_EQNINT      519
+#define CMD_EQNSLVi     520
+#define CMD_EQNINTi     521
+#define CMD_EQNVAR      522
+#define CMD_EQNMENU     523
+#define CMD_EQNMNU1     524
+#define CMD_SPPV        525
+#define CMD_SPFV        526
+#define CMD_USPV        527
+#define CMD_USFV        528
+#define CMD_GEN_N       529
+#define CMD_GEN_I       530
+#define CMD_GEN_PV      531
+#define CMD_GEN_PMT     532
+#define CMD_GEN_FV      533
+#define CMD_RAISE       534
+#define CMD_DIRECT      535
+#define CMD_NUMERIC     536
+#define CMD_EMBED       537
+#define CMD_NEWEQN      538
+#define CMD_EDITEQN     539
 /* Units */
-#define CMD_CONVERT     501
-#define CMD_UBASE       502
-#define CMD_UVAL        503
-#define CMD_UFACT       504
-#define CMD_TO_UNIT     505
-#define CMD_FROM_UNIT   506
-#define CMD_N_PLUS_U    507
-#define CMD_UNIT_T      508
+#define CMD_CONVERT     540
+#define CMD_UBASE       541
+#define CMD_UVAL        542
+#define CMD_UFACT       543
+#define CMD_TO_UNIT     544
+#define CMD_FROM_UNIT   545
+#define CMD_N_PLUS_U    546
+#define CMD_UNIT_T      547
 /* Directories */
-#define CMD_CRDIR       509
-#define CMD_PGDIR       510
-#define CMD_RENAME      511
-#define CMD_CHDIR       512
-#define CMD_UPDIR       513
-#define CMD_HOME        514
-#define CMD_PATH        515
-#define CMD_REFMOVE     516
-#define CMD_REFCOPY     517
-#define CMD_REFFIND     518
-#define CMD_PRALL       519
+#define CMD_CRDIR       548
+#define CMD_PGDIR       549
+#define CMD_RENAME      550
+#define CMD_CHDIR       551
+#define CMD_UPDIR       552
+#define CMD_HOME        553
+#define CMD_PATH        554
+#define CMD_REFMOVE     555
+#define CMD_REFCOPY     556
+#define CMD_REFFIND     557
+#define CMD_PRALL       558
 /* Big Screen */
-#define CMD_WIDTH       520
-#define CMD_HEIGHT      521
-#define CMD_HEADER      522
-#define CMD_ROW_PLUS    523
-#define CMD_ROW_MINUS   524
-#define CMD_COL_PLUS    525
-#define CMD_COL_MINUS   526
-#define CMD_GETDS       527
-#define CMD_SETDS       528
+#define CMD_WIDTH       559
+#define CMD_HEIGHT      560
+#define CMD_HEADER      561
+#define CMD_ROW_PLUS    562
+#define CMD_ROW_MINUS   563
+#define CMD_COL_PLUS    564
+#define CMD_COL_MINUS   565
+#define CMD_GETDS       566
+#define CMD_SETDS       567
+#define CMD_1LINE       568
+#define CMD_NLINE       569
+#define CMD_LTOP        570
+#define CMD_ATOP        571
+#define CMD_HFLAGS      572
+#define CMD_HPOLAR      573
+#define CMD_STK         574
 /* Menus */
-#define CMD_DIRS        529
-#define CMD_DIR_FCN     530
-#define CMD_UNITS       531
-#define CMD_UNIT_FCN    532
-#define CMD_PLOT_M      533
-#define CMD_TVM         534
-#define CMD_EQN         535
-#define CMD_EQN_FCN     536
-#define CMD_N           537
-#define CMD_I_PCT_YR    538
-#define CMD_PV          539
-#define CMD_PMT         540
-#define CMD_FV          541
-#define CMD_P_PER_YR    542
-#define CMD_TBEGIN      543
-#define CMD_TEND        544
-#define CMD_TCLEAR      545
-#define CMD_TRESET      546
-#define CMD_AMORT       547
-#define CMD_TNUM_P      548
-#define CMD_TINT        549
-#define CMD_TPRIN       550
-#define CMD_TBAL        551
-#define CMD_TNEXT       552
-#define CMD_TFIRST      553
-#define CMD_TLAST       554
-#define CMD_TINCR       555
-#define CMD_TGO         556
+#define CMD_DIRS        575
+#define CMD_DIR_FCN     576
+#define CMD_UNITS       577
+#define CMD_UNIT_FCN    578
+#define CMD_PLOT_M      579
+#define CMD_TVM         580
+#define CMD_EQN         581
+#define CMD_EQN_FCN     582
+#define CMD_N           583
+#define CMD_I_PCT_YR    584
+#define CMD_PV          585
+#define CMD_PMT         586
+#define CMD_FV          587
+#define CMD_P_PER_YR    588
+#define CMD_TBEGIN      589
+#define CMD_TEND        590
+#define CMD_TCLEAR      591
+#define CMD_TRESET      592
+#define CMD_AMORT       593
+#define CMD_TNUM_P      594
+#define CMD_TINT        595
+#define CMD_TPRIN       596
+#define CMD_TBAL        597
+#define CMD_TNEXT       598
+#define CMD_TFIRST      599
+#define CMD_TLAST       600
+#define CMD_TINCR       601
+#define CMD_TGO         602
 /* Graphing */
-#define CMD_PGMPLOT     557
-#define CMD_EQNPLOT     558
-#define CMD_PARAM       559
-#define CMD_XAXIS       560
-#define CMD_YAXIS       561
-#define CMD_CONST       562
-#define CMD_VIEW_P      563
-#define CMD_XMIN        564
-#define CMD_XMAX        565
-#define CMD_YMIN        566
-#define CMD_YMAX        567
-#define CMD_SCAN        568
-#define CMD_PLOT        569
-#define CMD_LINE        570
-#define CMD_LIFE        571
+#define CMD_PGMPLOT     603
+#define CMD_EQNPLOT     604
+#define CMD_PARAM       605
+#define CMD_XAXIS       606
+#define CMD_YAXIS       607
+#define CMD_CONST       608
+#define CMD_VIEW_P      609
+#define CMD_XMIN        610
+#define CMD_XMAX        611
+#define CMD_YMIN        612
+#define CMD_YMAX        613
+#define CMD_SCAN        614
+#define CMD_PLOT        615
+#define CMD_LINE        616
+#define CMD_LIFE        617
 
-#define CMD_SENTINEL    572
+#define CMD_SENTINEL    618
 
 
 /* command_spec.argtype */
@@ -645,10 +692,12 @@
 #define ARG_PRGM     12 /* Alpha label (CATSECT_PGM) */
 #define ARG_RVAR     13 /* Variable (real only) (MVAR, INTEG, SOLVE) */
 #define ARG_MAT      14 /* Variable (matrix only) (EDITN, INDEX) */
-#define ARG_EQN      15 /* Variable (equation only) (EVAL) */
-#define ARG_DIR      16 /* Directory name (CRDIR, PGDIR, CHDIR) */
-#define ARG_XSTR     17 /* Long string (XSTR) */
-#define ARG_OTHER    18 /* Weirdos */
+#define ARG_M_STK    15 /* Matrix variable or stack (GETMI, PUTMI) */
+#define ARG_L_STK    16 /* List variable or stack (GETLI, PUTLI) */
+#define ARG_EQN      17 /* Variable (equation only) (EVAL) */
+#define ARG_DIR      18 /* Directory name (CRDIR, PGDIR, CHDIR) */
+#define ARG_XSTR     19 /* Long string (XSTR) */
+#define ARG_OTHER    20 /* Weirdos */
 
 
 /* command_spec.flags */
